@@ -19,7 +19,7 @@
 #include <linux/netdevice.h>
 #include <linux/etherdevice.h>
 #include <linux/mm.h>
-#include <linux/skbuff.h>
+#include "compat_skbuff.h"
 #include <linux/sockios.h>
 #include "compat_sock.h"
 
@@ -72,12 +72,15 @@ struct VNetBridge {
    VNetPort                 port;           // connection to virtual hub
    Bool                     wirelessAdapter; // connected to wireless adapter?
    struct SMACState        *smac;           // device structure for wireless
+#ifdef VMW_NETDEV_HAS_NET
+   struct net_device       *internalDev;
+#endif
 };
 
 typedef PacketStatus (* SMACINT SMACFunc)(struct SMACState *, SMACPackets *);
 
-static int  VNetBridgeUp(VNetBridge *bridge);
-static void VNetBridgeDown(VNetBridge *bridge);
+static int  VNetBridgeUp(VNetBridge *bridge, Bool rtnlLock);
+static void VNetBridgeDown(VNetBridge *bridge, Bool rtnlLock);
 
 static int  VNetBridgeNotify(struct notifier_block *this, u_long msg,
 			     void *data);
@@ -109,9 +112,7 @@ static int  VNetBridgeProcRead(char *page, char **start, off_t off,
  *
  *      Set IFF_PROMISC on the peer interface.
  *
- *      XXX need locking?
- *
- * Results: 
+ * Results:
  *      None.
  *
  * Side effects:
@@ -121,15 +122,19 @@ static int  VNetBridgeProcRead(char *page, char **start, off_t off,
  */
 
 static void
-VNetBridgeStartPromisc(VNetBridge *bridge)
+VNetBridgeStartPromisc(VNetBridge *bridge,      // IN:
+                       Bool rtnlLock)           // IN: Acquire RTNL lock
 {
    struct net_device *dev = bridge->dev;
 
-   /* 
+   /*
     * Disable wireless cards from going into promiscous mode because those
     * cards which do support RF monitoring would not be able to function
     * correctly i.e. they would not be able to send data packets.
     */
+   if (rtnlLock) {
+      rtnl_lock();
+   }
    if (!bridge->enabledPromisc && !bridge->wirelessAdapter) {
       dev_set_promiscuity(dev, 1);
       bridge->enabledPromisc = TRUE;
@@ -137,6 +142,9 @@ VNetBridgeStartPromisc(VNetBridge *bridge)
       LOG(0, (KERN_NOTICE "bridge-%s: enabled promiscuous mode\n",
 	      bridge->name));
    }
+   if (rtnlLock) {
+      rtnl_unlock();
+   }
 }
 
 
@@ -147,9 +155,7 @@ VNetBridgeStartPromisc(VNetBridge *bridge)
  *
  *      Restore saved IFF_PROMISC on the peer interface.
  *
- *      XXX need locking?
- *
- * Results: 
+ * Results:
  *      None.
  *
  * Side effects:
@@ -159,16 +165,23 @@ VNetBridgeStartPromisc(VNetBridge *bridge)
  */
 
 static void
-VNetBridgeStopPromisc(VNetBridge *bridge)
+VNetBridgeStopPromisc(VNetBridge *bridge,       // IN:
+                      Bool rtnlLock)            // IN: Acquire RTNL lock
 {
    struct net_device *dev = bridge->dev;
 
+   if (rtnlLock) {
+      rtnl_lock();
+   }
    if (bridge->enabledPromisc && !bridge->wirelessAdapter) {
       dev_set_promiscuity(dev, -1);
-      bridge->enabledPromisc = FALSE;   
+      bridge->enabledPromisc = FALSE;
       LOG(0, (KERN_NOTICE "bridge-%s: disabled promiscuous mode\n",
 	      bridge->name));
    }
+   if (rtnlLock) {
+      rtnl_unlock();
+   }
 }
 
 
@@ -181,9 +194,7 @@ VNetBridgeStopPromisc(VNetBridge *bridge)
  *
  *      This can be called periodically.
  *
- *      XXX need locking?
- *
- * Results: 
+ * Results:
  *      None.
  *
  * Side effects:
@@ -195,6 +206,7 @@ VNetBridgeStopPromisc(VNetBridge *bridge)
 static INLINE_SINGLE_CALLER void
 VNetBridgeCheckPromisc(VNetBridge *bridge)
 {
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
    if (bridge->enabledPromisc && !bridge->wirelessAdapter) {
       struct net_device *dev = bridge->dev;
       Bool devPromisc = (dev->flags & IFF_PROMISC) != 0;
@@ -206,9 +218,67 @@ VNetBridgeCheckPromisc(VNetBridge *bridge)
 	            "Your Ethernet driver is not compatible with VMware's bridged networking.\n",
                     bridge->name));
          }
+         rtnl_lock();
          dev_set_promiscuity(dev, 0);
+         rtnl_unlock();
       }
    }
+#endif
+}
+
+
+#ifdef VMW_NETDEV_HAS_NET
+/*
+ *----------------------------------------------------------------------
+ *
+ * VNetBridgeInternalSetup --
+ *
+ *      Setup callback for our bridge internal device.  Nothing to do,
+ *      generic code sets up everything we expect from device.
+ *
+ * Results:
+ *      None.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static void
+VNetBridgeInternalSetup(struct net_device *net)
+{
+   /* Do nothing. */
+}
+#endif
+
+
+/*
+ *----------------------------------------------------------------------
+ *
+ * VNetBridgeDevCompatible --
+ *
+ *      Check whether bridge and network device are compatible.
+ *
+ * Results:
+ *      Non-zero if device is good enough for bridge.  Zero otherwise.
+ *
+ * Side effects:
+ *      None.
+ *
+ *----------------------------------------------------------------------
+ */
+
+static INLINE_SINGLE_CALLER int
+VNetBridgeDevCompatible(VNetBridge *bridge,      // IN: Bridge
+                        struct net_device *net)  // IN: Network device
+{
+#ifdef VMW_NETDEV_HAS_NET
+   if (net->nd_net != bridge->internalDev->nd_net) {
+      return 0;
+   }
+#endif
+   return strcmp(net->name, bridge->name) == 0;
 }
 
 
@@ -264,6 +334,14 @@ VNetBridge_Create(char      *devName, // IN:  name of device (e.g., "eth0")
    memcpy(bridge->name, devName, sizeof bridge->name);
    NULL_TERMINATE_STRING(bridge->name);
 
+#ifdef VMW_NETDEV_HAS_NET
+   bridge->internalDev = compat_alloc_netdev(0, "vmnetX", VNetBridgeInternalSetup);
+   if (!bridge->internalDev) {
+      retval = -ENOMEM;
+      goto out;
+   }
+#endif
+
    /*
     * Set up notifier for network device state change
     */
@@ -276,7 +354,7 @@ VNetBridge_Create(char      *devName, // IN:  name of device (e.g., "eth0")
     * Try to bring it up
     */
 
-   retval = VNetBridgeUp(bridge);
+   retval = VNetBridgeUp(bridge, TRUE);
    if (retval == -ENODEV) {
       LOG(1, (KERN_DEBUG "bridge-%s: peer interface %s not found, "
 	      "will wait for it to come up\n",
@@ -352,6 +430,11 @@ out:
       if (bridge->notifier.notifier_call != NULL) {
          unregister_netdevice_notifier(&bridge->notifier);
       }
+#ifdef VMW_NETDEV_HAS_NET
+      if (bridge->internalDev) {
+         compat_free_netdev(bridge->internalDev);
+      }
+#endif
       kfree(bridge);
    }
    return retval;
@@ -381,11 +464,17 @@ VNetBridgeFree(VNetJack *this) // IN: jack to free
    VNetBridge *bridge = (VNetBridge*)this->private;
    
    if (bridge->dev != NULL) {
-      VNetBridgeDown(bridge);
+      VNetBridgeDown(bridge, TRUE);
    }
 
    unregister_netdevice_notifier(&bridge->notifier);
 
+#ifdef VMW_NETDEV_HAS_NET
+   if (bridge->internalDev) {
+      compat_free_netdev(bridge->internalDev);
+   }
+#endif
+
    if (this->procEntry) {
       VNetProc_RemoveEntry(this->procEntry, NULL);
    }
@@ -707,9 +796,9 @@ VNetBridgePortsChanged(VNetJack *this) // IN: jack
    VNetBridge *bridge = (VNetBridge*)this->private;
    if (bridge->dev) {
       if (VNetGetAttachedPorts(this)) {
-         VNetBridgeStartPromisc(bridge);
+         VNetBridgeStartPromisc(bridge, TRUE);
       } else {
-         VNetBridgeStopPromisc(bridge);
+         VNetBridgeStopPromisc(bridge, TRUE);
       }
    }
 }
@@ -795,7 +884,8 @@ VNetBridgeIsDeviceWireless(struct net_device *dev) //IN: sock
  */
 
 int
-VNetBridgeUp(VNetBridge *bridge) // IN: bridge struct
+VNetBridgeUp(VNetBridge *bridge, // IN: bridge struct
+             Bool rtnlLock)      // IN: acquire RTNL lock
 {
    int retval = 0;
 
@@ -809,7 +899,7 @@ VNetBridgeUp(VNetBridge *bridge) // IN: bridge struct
     */
 
    dev_lock_list();
-   bridge->dev = DEV_GET(bridge->name);
+   bridge->dev = DEV_GET(bridge);
    LOG(2, (KERN_DEBUG "bridge-%s: got dev %p\n",
 	   bridge->name, bridge->dev));
    if (bridge->dev == NULL) {
@@ -847,13 +937,13 @@ VNetBridgeUp(VNetBridge *bridge) // IN: bridge struct
     * (The alternative is to support the callback.)
     */
 
-   bridge->sk = compat_sk_alloc(GFP_ATOMIC);
+   bridge->sk = compat_sk_alloc(bridge, GFP_ATOMIC);
    if (bridge->sk == NULL) {
       dev_unlock_list();
       retval = -ENOMEM;
       goto out;
    }
-   SET_SK_DEAD(bridge->sk, 1);
+   SET_SK_DEAD(bridge->sk);
 
    bridge->wirelessAdapter = VNetBridgeIsDeviceWireless(bridge->dev);
 
@@ -904,7 +994,7 @@ VNetBridgeUp(VNetBridge *bridge) // IN: bridge struct
 
    down(&vnetStructureSemaphore);
    if (VNetGetAttachedPorts(&bridge->port.jack)) {
-      VNetBridgeStartPromisc(bridge);
+      VNetBridgeStartPromisc(bridge, rtnlLock);
    }
    up(&vnetStructureSemaphore);
    
@@ -949,14 +1039,15 @@ out:
  */
 
 void
-VNetBridgeDown(VNetBridge *bridge) // IN: bridge
+VNetBridgeDown(VNetBridge *bridge, // IN: bridge
+               Bool rtnlLock)      // IN: acquire RTNL lock
 {
    if (bridge->dev == NULL) {
       LOG(0, (KERN_NOTICE "bridge-%s: already down\n", bridge->name));
       return;
    }
 
-   VNetBridgeStopPromisc(bridge);
+   VNetBridgeStopPromisc(bridge, rtnlLock);
    if (bridge->smac){
       SMAC_SetMac(bridge->smac, NULL);
    }
@@ -1002,7 +1093,7 @@ VNetBridgeNotify(struct notifier_block *this, // IN: callback data (bridge)
          LOG(0, (KERN_WARNING "bridge-%s: interface %s unregistered without "
                  "going down! Disabling the bridge\n", bridge->name,
                  dev->name));
-         VNetBridgeDown(bridge);
+         VNetBridgeDown(bridge, FALSE);
       }
       break;
 
@@ -1012,18 +1103,18 @@ VNetBridgeNotify(struct notifier_block *this, // IN: callback data (bridge)
       if (dev == bridge->dev) {
          LOG(1, (KERN_DEBUG "bridge-%s: disabling the bridge\n",
                  bridge->name));
-         VNetBridgeDown(bridge);
+         VNetBridgeDown(bridge, FALSE);
       }
       break;
 
    case NETDEV_UP:
       LOG(2, (KERN_DEBUG "bridge-%s: interface %s is going up\n",
               bridge->name, dev->name));
-      if (bridge->dev == NULL && strcmp(dev->name, bridge->name) == 0) {
+      if (bridge->dev == NULL && VNetBridgeDevCompatible(bridge, dev)) {
          int errno;
 
          LOG(1, (KERN_DEBUG "bridge-%s: enabling the bridge\n", bridge->name));
-	 errno = VNetBridgeUp(bridge);
+	 errno = VNetBridgeUp(bridge, FALSE);
 	 switch (-errno) {
 	 case 0:
 	    break;
@@ -1074,7 +1165,7 @@ VNetBridgeNotify(struct notifier_block *this, // IN: callback data (bridge)
  *      None.
  *
  * Side effects:
- *      skb->h.raw updated to point to the tcp/udp header.
+ *      transport header pointer updated to point to the tcp/udp header.
  *
  *----------------------------------------------------------------------
  */
@@ -1083,16 +1174,20 @@ static INLINE_SINGLE_CALLER void
 VNetBridgeComputeHeaderPos(struct sk_buff *skb) // IN: buffer to examine
 {
    /* Maybe some kernel gets it right... */
-   if (skb->h.raw != skb->nh.raw) {
+   if (compat_skb_network_header_len(skb)) {
       return;
    }
    switch (be16_to_cpu(skb->protocol)) {
-      case ETH_P_IP:
-         skb->h.raw = skb->nh.raw + (skb->nh.raw[0] & 0x0F) * 4;
+      case ETH_P_IP: {
+            struct iphdr *ipHdr = compat_skb_ip_header(skb);
+
+            compat_skb_set_transport_header(skb, compat_skb_network_offset(skb) +
+                                                 ipHdr->ihl * 4);
+         }
 	 return;
       default:
          LOG(3, (KERN_DEBUG "Unknown EII protocol %04X: csum at %d\n", 
-		 be16_to_cpu(skb->protocol), skb->csum));
+		 be16_to_cpu(skb->protocol), compat_skb_csum_offset(skb)));
 	 break;
    }
    return;
@@ -1143,21 +1238,21 @@ VNetBridgeGSOSegment(struct sk_buff *skb)        // IN: packet to split
    uint16 ipID;
    uint32 seqNo;
 
-   if (((struct ethhdr *)skb->mac.raw)->h_proto != htons(ETH_P_IP)) {
+   if (((struct ethhdr *)compat_skb_mac_header(skb))->h_proto != htons(ETH_P_IP)) {
       return ERR_PTR(-EPFNOSUPPORT);
    }
 
-   if (skb->nh.iph->protocol != IPPROTO_TCP) {
+   if (compat_skb_ip_header(skb)->protocol != IPPROTO_TCP) {
       return ERR_PTR(-EPROTONOSUPPORT);
    }
 
-   macHdrLen = skb->nh.raw - skb->mac.raw;
-   ipHdrLen = skb->nh.iph->ihl << 2;
-   tcpHdrLen = skb->h.th->doff << 2;
+   macHdrLen = compat_skb_network_header(skb) - compat_skb_mac_header(skb);
+   ipHdrLen = compat_skb_ip_header(skb)->ihl << 2;
+   tcpHdrLen = compat_skb_tcp_header(skb)->doff << 2;
    allHdrLen = macHdrLen + ipHdrLen + tcpHdrLen;
 
-   ipID = ntohs(skb->nh.iph->id);
-   seqNo = ntohl(skb->h.th->seq);
+   ipID = ntohs(compat_skb_ip_header(skb)->id);
+   seqNo = ntohl(compat_skb_tcp_header(skb)->seq);
 
    /* Host TCP stack populated this (MSS) for the host NIC driver */
    bytesPerPacket = skb_shinfo(skb)->tso_size;
@@ -1194,36 +1289,37 @@ VNetBridgeGSOSegment(struct sk_buff *skb)        // IN: packet to split
       memcpy(skb_put(newSkb, allHdrLen), skb->data, allHdrLen);
 
       /* Fix up pointers to different layers */
-      newSkb->mac.raw = newSkb->data;
-      newSkb->nh.raw = newSkb->data + macHdrLen;
-      newSkb->h.raw = newSkb->nh.raw + ipHdrLen;
+      compat_skb_reset_mac_header(newSkb);
+      compat_skb_set_network_header(newSkb, macHdrLen);
+      compat_skb_set_transport_header(newSkb, macHdrLen + ipHdrLen);
 
       /* Payload copy */
-      skb_copy_bits(skb, curByteOffset, newSkb->tail, payloadSize);
+      skb_copy_bits(skb, curByteOffset, compat_skb_tail_pointer(newSkb), payloadSize);
       skb_put(newSkb, payloadSize);
 
       curByteOffset+=payloadSize;
       bytesLeft -= payloadSize;
 
       /* Fix up IP hdr */
-      newSkb->nh.iph->tot_len = htons(payloadSize + tcpHdrLen + ipHdrLen);
-      newSkb->nh.iph->id = htons(ipID);
-      newSkb->nh.iph->check = 0;
+      compat_skb_ip_header(newSkb)->tot_len = htons(payloadSize + tcpHdrLen + ipHdrLen);
+      compat_skb_ip_header(newSkb)->id = htons(ipID);
+      compat_skb_ip_header(newSkb)->check = 0;
       /* Recompute new IP checksum */
-      newSkb->nh.iph->check =
-         ip_fast_csum((uint8 *)newSkb->nh.iph, newSkb->nh.iph->ihl);
+      compat_skb_ip_header(newSkb)->check =
+              ip_fast_csum(compat_skb_network_header(newSkb),
+                           compat_skb_ip_header(newSkb)->ihl);
 
       /* Fix up TCP hdr */
-      newSkb->h.th->seq = htonl(seqNo);
+      compat_skb_tcp_header(newSkb)->seq = htonl(seqNo);
       /* Clear FIN/PSH if not last packet */
       if (bytesLeft > 0) {
-         newSkb->h.th->fin = 0;
-         newSkb->h.th->psh = 0;
+         compat_skb_tcp_header(newSkb)->fin = 0;
+         compat_skb_tcp_header(newSkb)->psh = 0;
       }
       /* Recompute partial TCP checksum */
-      newSkb->h.th->check =
-         ~csum_tcpudp_magic(newSkb->nh.iph->saddr,
-                            newSkb->nh.iph->daddr,
+      compat_skb_tcp_header(newSkb)->check =
+         ~csum_tcpudp_magic(compat_skb_ip_header(newSkb)->saddr,
+                            compat_skb_ip_header(newSkb)->daddr,
                             payloadSize+tcpHdrLen, IPPROTO_TCP, 0);
 
       /* Offset of field */
@@ -1370,7 +1466,7 @@ VNetBridgeReceiveFromDev(struct sk_buff *skb,         // IN: packet to receive
 #  endif
 
    if (bridge->smac) {
-      if (VNetCallSMACFunc(bridge->smac, &skb, skb->mac.raw, 
+      if (VNetCallSMACFunc(bridge->smac, &skb, compat_skb_mac_header(skb), 
                            SMAC_CheckPacketFromHost) != 
           PacketStatusForwardPacket) {
          LOG(4, (KERN_NOTICE "bridge-%s: packet dropped .\n",
@@ -1392,7 +1488,7 @@ VNetBridgeReceiveFromDev(struct sk_buff *skb,         // IN: packet to receive
 #endif
 #endif
    
-   skb_push(skb, skb->data - skb->mac.raw);
+   skb_push(skb, skb->data - compat_skb_mac_header(skb));
    LOG(3, (KERN_DEBUG "bridge-%s: receive %d\n",
 	   bridge->name, (int) skb->len));
    if (VNetBridgeIsGSO(skb)) {
